/*
 ** fbcollector.c
 ** IPFIX Collecting Process single transport session implementation
 **
 ** ------------------------------------------------------------------------
 ** Copyright (C) 2006-2008 Carnegie Mellon University. All Rights Reserved.
 ** ------------------------------------------------------------------------
 ** Authors: Brian Trammell <bht@cert.org>
 ** ------------------------------------------------------------------------
 ** GNU Lesser GPL Rights pursuant to Version 2.1, February 1999 
 ** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
 ** ------------------------------------------------------------------------
 */

#define _FIXBUF_SOURCE_
#include <fixbuf/private.h>

typedef gboolean    (*fbCollectorRead_fn) (
    fbCollector_t               *collector,
    uint8_t                     *msgbase,
    size_t                      *msglen,
    GError                      **err);

typedef void        (*fbCollectorClose_fn)(
    fbCollector_t               *collector);

struct fbCollector_st {
    /** Listener from which this Collector was created. */
    fbListener_t                *listener;
    /** 
     * Application context. Created and owned by the application
     * when the listener calls the <new collector> callback.
     */
    void                        *ctx;
    /** Cached peer address. Filled in at allocation time */
    union {
        struct sockaddr         so;
        struct sockaddr_in      ip4;
        struct sockaddr_in6     ip6;
    }                           peer;
    /** Current export stream */
    union {
        /** Buffered file pointer, for file transport */
        FILE                    *fp;
        /** 
         * Unbuffered socket, for SCTP, TCP, or UDP transport.
         * Also used as base socket for TLS and DTLS support.
         */
        int                     fd;
    }                           stream;
#if HAVE_OPENSSL
    /** OpenSSL socket, for TLS or DTLS over the socket in fd. */
    SSL                         *ssl;
#endif
    gboolean                    active;
    fbCollectorRead_fn          coread;
    fbCollectorClose_fn         coclose;
};

typedef struct fbCollectorMsgVL_st {
    uint16_t                    n_version;
    uint16_t                    n_len;
} fbCollectorMsgVL_t;

static uint16_t fbCollectorDecodeMsgVL(
    fbCollectorMsgVL_t          *hdr,
    size_t                      b_len,
    GError                      **err)
{
    uint16_t                    h_version;
    uint16_t                    h_len;
    
    h_version = g_ntohs(hdr->n_version);
    if (h_version != 0x000A) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                    "Illegal IPFIX Message version 0x%04x; "
                    "input is probably not an IPFIX Message stream.",
                    g_ntohs(hdr->n_version));
        return 0;
    }
    
    h_len = g_ntohs(hdr->n_len);
    if (h_len < 16) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                    "Illegal IPFIX Message length 0x%04x; "
                    "input is probably not an IPFIX Message stream.",
                    g_ntohs(hdr->n_len));
        return 0;
    }
    
    if (b_len && (h_len > b_len)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_BUFSZ,
                    "Buffer too small to read IPFIX Message "
                    "(message size %hu, buffer size %lu).",
                    h_len, b_len);
        return 0;
    }
    
    return h_len;
}

static gboolean fbCollectorReadFile(
    fbCollector_t           *collector,
    uint8_t                 *msgbase,
    size_t                  *msglen,
    GError                  **err)
{
    int                     rc;
    uint16_t                h_len;
    
    /* Read and decode version and length */
    g_assert(*msglen > 4);
    rc = fread(msgbase, 1, 4, collector->stream.fp);
    if (rc > 0) {
        h_len = fbCollectorDecodeMsgVL((fbCollectorMsgVL_t *)msgbase, 
                                        *msglen, err);
        if (!h_len) return FALSE;
        msgbase += 4;
    } else if (feof(collector->stream.fp)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                    "End of file");
        return FALSE;
    } else {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "I/O error: %s", strerror(errno));
        return FALSE;
    }

    /* read rest of message */
    rc = fread(msgbase, 1, h_len - 4, collector->stream.fp);
    if (rc > 0) {
        *msglen = rc + 4;
        return TRUE;
    } else if (feof(collector->stream.fp)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                    "End of file");
        return FALSE;
    } else {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "I/O error: %s", strerror(errno));
        return FALSE;
    }
}

static void fbCollectorCloseFile(
    fbCollector_t   *collector)
{
    if (collector->stream.fp != stdin) {
        fclose(collector->stream.fp);
    }
    collector->active = FALSE;
}

fbCollector_t *fbCollectorAllocFP(
    void            *ctx,
    FILE            *fp)
{
    fbCollector_t   *collector = NULL;
    
    /* Create a new collector */
    collector = g_slice_new0(fbCollector_t);
    
    /* Fill the collector in */
    collector->ctx = ctx;
    collector->stream.fp = fp;
    collector->active = TRUE;
    collector->coread = fbCollectorReadFile;
    
    /* All done */
    return collector;
}

fbCollector_t *fbCollectorAllocFile(
    void            *ctx,
    const char      *path,
    GError          **err)
{
    fbCollector_t   *collector = NULL;
    FILE            *fp = NULL;

    /* check to see if we're opening stdin */
    if ((strlen(path) == 1) && (path[0] == '-')) 
    {
        /* don't open a terminal */
        if (isatty(fileno(stdin))) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "Refusing to open stdin terminal for collection");
            return NULL;
        }
        
        /* yep, stdin */
        fp = stdin;
    } else {
        /* nope, just a regular file; open it. */
        fp = fopen(path, "r");
    }
    
    /* check for error */
    if (!fp) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "Couldn't open %s for collection: %s",
                    path, strerror(errno));
        return NULL;
    }
    
    /* allocate a collector */
    collector = fbCollectorAllocFP(ctx, fp);
    
    /* set the file close function */
    collector->coclose = fbCollectorCloseFile;
    
    /* all done */
    return collector;
}

#if FB_ENABLE_SCTP

static gboolean fbCollectorReadSCTP(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    struct sockaddr         peer;
    socklen_t               peerlen = sizeof(peer);
    struct sctp_sndrcvinfo  sri;
    int                     sctp_flags = 0;
    int                     rc;
    
    rc = sctp_recvmsg(collector->stream.fd, msgbase, *msglen,
                      &peer, &peerlen, &sri, &sctp_flags);
    
    if (rc > 0) {
        *msglen = rc;
        return TRUE;
    } else if (rc == 0) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                    "End of file");
        return FALSE;
    } else if (errno == EINTR) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                    "SCTP read interrupt");
        return FALSE;
    } else {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "TCP I/O error: %s", strerror(errno));
        return FALSE;
    }
}
#endif /* FB_ENABLE_SCTP */

static gboolean fbCollectorReadTCP(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    int                     rc;
    uint16_t                h_len, rrem;
    
    /* Read and decode version and length */
    g_assert(*msglen > 4);
    rrem = 4;
    while (rrem) {
        rc = read(collector->stream.fd, msgbase, rrem);
        if (rc > 0) {
            rrem -= rc;
            msgbase += rc;
        } else if (rc == 0) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                        "End of file");
            return FALSE;
        } else if (errno == EINTR) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                        "TCP read interrupt at message start");
            return FALSE;
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "TCP I/O error: %s", strerror(errno));
            return FALSE;
        }
    }
    h_len = fbCollectorDecodeMsgVL((fbCollectorMsgVL_t *)(msgbase - 4),
                                   *msglen, err);
    if (!h_len) return FALSE;
    
    /* read rest of message */
    rrem = h_len - 4;
    while (rrem) {
        rc = read(collector->stream.fd, msgbase, rrem);
        if (rc > 0) {
            rrem -= rc;
            msgbase += rc;
        } else if (rc == 0) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                        "End of file");
            return FALSE;
        } else if (errno == EINTR) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                        "TCP read interrupt in message");
            return FALSE;
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "TCP I/O error: %s", strerror(errno));
            return FALSE;
        }
    }
    
    /* All done. Return message length from header. */
    *msglen = h_len;
    return TRUE;
}

static gboolean fbCollectorVerifyUDPPeer(
    fbCollector_t   *collector,
    struct sockaddr *from,
    socklen_t       fromlen,
    GError          **err)
{
    /* stash the address if we've not seen it before */
    /* compare the address if we have */
    /* simulate no data (NLREAD) if message is from wrong peer */
    /* for now always verify */
    return TRUE;
}

static gboolean fbCollectorReadUDP(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    ssize_t         recvlen;
    union {
        struct sockaddr         so;
        struct sockaddr_in      ip4;
        struct sockaddr_in6     ip6;
    }                           peer;
    socklen_t                   peerlen;

    peerlen = sizeof(peer);
    recvlen = recvfrom(collector->stream.fd, msgbase, *msglen, 0,
                       (struct sockaddr *)&peer, &peerlen);
    
    if (recvlen > 0) {
        *msglen = recvlen;
        /* 
         * FIXME this is a bit broken. What we're doing here is
         * ensuring we reject any message not from the _first_ peer
         * (thereby ensuring session integrity). To really do a UDP
         * Collecting Process _right_ (to the extent that such is
         * possible, heh...), we'd need to map peer addresses to
         * sessions, not sockets to CP sessions.
         *
         * Seems a lot of work for an insecure optional protocol.
         * So we do this, instead.
         */
        return fbCollectorVerifyUDPPeer(collector, &(peer.so), peerlen, err);
    } else if (errno == EINTR || errno == EWOULDBLOCK) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                        "UDP read interrupt or timeout");
        return FALSE;
    } else {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "UDP I/O error: %s", strerror(errno));
        return FALSE;
    }
}

static void fbCollectorCloseSocket(
    fbCollector_t   *collector)
{
    close(collector->stream.fd);
    collector->active = FALSE;
}

fbCollector_t *fbCollectorAllocSocket(
    fbListener_t    *listener,
    void            *ctx,
    int             fd,
    struct sockaddr *peer,
    size_t          peerlen)
{
    fbCollector_t   *collector = NULL;
    fbConnSpec_t    *spec = fbListenerGetConnSpec(listener);
    
    /* Create a new collector */
    collector = g_slice_new0(fbCollector_t);
    
    /* Fill it in */
    collector->listener = listener;
    collector->ctx = ctx;
    collector->stream.fd = fd;
    collector->active = TRUE;
    collector->coclose = fbCollectorCloseSocket;
    if (peerlen) {
        memcpy(&(collector->peer.so), peer, 
               (peerlen > sizeof(collector->peer)) ? 
                    sizeof(collector->peer) : peerlen);
    }

    /* Select a reader function */
    switch(spec->transport) {
#if FB_ENABLE_SCTP
    case FB_SCTP:
        collector->coread = fbCollectorReadSCTP;
        break;
#endif
    case FB_TCP:
        collector->coread = fbCollectorReadTCP;
        break;
    case FB_UDP:
        collector->coread = fbCollectorReadUDP;
        break;
    default:
        g_assert_not_reached();
    }
    
    /* All done */
    return collector;
}

#if HAVE_OPENSSL

static gboolean fbCollectorReadTLS(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    int                     rc;
    uint16_t                h_len, rrem;
    
    /* Read and decode version and length */
    g_assert(*msglen > 4);
    rrem = 4;
    while (rrem) {
        rc = SSL_read(collector->ssl, msgbase, rrem);
        if (rc > 0) {
            rrem -= rc;
            msgbase += rc;
        } else if (rc == 0) {
            /* FIXME this isn't _quite_ robust but it's good enough for now.
               we'll fix this when we do TLS/TCP stress testing. */
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                        "TLS connection shutdown");
            return FALSE;
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "TLS I/O error at message start: %s",
                        ERR_error_string(ERR_get_error(), NULL));
            while (ERR_get_error());
            return FALSE;
        }
    }
    h_len = fbCollectorDecodeMsgVL((fbCollectorMsgVL_t *)(msgbase - 4),
                                   *msglen, err);
    if (!h_len) return FALSE;
    
    /* read rest of message */
    rrem = h_len - 4;
    while (rrem) {
        rc = SSL_read(collector->ssl, msgbase, rrem);
        if (rc > 0) {
            rrem -= rc;
            msgbase += rc;
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "TLS I/O error in message: %s",
                        ERR_error_string(ERR_get_error(), NULL));
            while (ERR_get_error());
            return FALSE;
        }
    }
    
    /* All done. Return message length from header. */
    *msglen = h_len;
    return TRUE;
}

static void fbCollectorCloseTLS(
    fbCollector_t   *collector)
{
    SSL_shutdown(collector->ssl);
    SSL_free(collector->ssl);
    collector->active = FALSE;
}

static gboolean fbCollectorOpenTLS(
    fbCollector_t   *collector,
    GError          **err)
{
    fbConnSpec_t    *spec = fbListenerGetConnSpec(collector->listener);
    BIO             *conn;
    gboolean        ok = TRUE;
    
    /* Initialize SSL context if necessary */
    if (!spec->vssl_ctx) {
        if (!fbConnSpecInitTLS(spec, TRUE, err)) {
            return FALSE;
        }
    }
   
    /* wrap a stream BIO around the opened socket */
    if (!(conn = BIO_new_socket(collector->stream.fd, 1))) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldn't wrap socket for TLS: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;
    }
    
    /* create SSL socket */
    if (!(collector->ssl = SSL_new((SSL_CTX *)spec->vssl_ctx))) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldnt create TLS socket: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;
    }
    
    /* accept SSL connection */
    SSL_set_accept_state(collector->ssl);
    SSL_set_bio(collector->ssl, conn, conn);
    SSL_set_mode(collector->ssl, SSL_MODE_AUTO_RETRY);   
    if (SSL_accept(collector->ssl) <= 0) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldn't accept on connected TLS socket: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;    
    }
    
    /* FIXME do post-connection verification */
    
end:
    if (!ok) {
        collector->active = FALSE;
        if (collector->ssl) {
            SSL_free(collector->ssl);
            collector->ssl = NULL;
        } else if (conn) {
            BIO_vfree(conn);
        }
    }
    return ok;
}

#if HAVE_OPENSSL_DTLS

static gboolean fbCollectorOpenDTLS(
    fbCollector_t   *collector,
    GError          **err)
{
    fbConnSpec_t    *spec = fbListenerGetConnSpec(collector->listener);
    BIO             *conn;
    gboolean        ok = TRUE;
    
    /* Initialize SSL context if necessary */
    if (!spec->vssl_ctx) {
        if (!fbConnSpecInitTLS(spec, TRUE, err)) {
            return FALSE;
        }
    }
    
    /* wrap a stream BIO around the opened socket */
    if (!(conn = BIO_new_dgram(collector->stream.fd, 1))) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldn't wrap socket for TLS: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;
    }
    
    /* create SSL socket */
    if (!(collector->ssl = SSL_new((SSL_CTX *)spec->vssl_ctx))) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldnt create TLS socket: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;
    }
    
    /* Enable cookie exchange */
    SSL_set_options(collector->ssl, SSL_OP_COOKIE_EXCHANGE);

    /* accept SSL connection */
    SSL_set_bio(collector->ssl, conn, conn);
    SSL_set_accept_state(collector->ssl);
    SSL_set_mode(collector->ssl, SSL_MODE_AUTO_RETRY);   
    if (SSL_accept(collector->ssl) <= 0) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldn't accept on connected TLS socket: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;    
    }

    /* FIXME do post-connection verification */
    
end:
    if (!ok) {
        collector->active = FALSE;
        if (collector->ssl) {
            SSL_free(collector->ssl);
            collector->ssl = NULL;
        } else if (conn) {
            BIO_vfree(conn);
        }
    }
    return ok;
}

#endif /* HAVE_OPENSSL_DTLS */

fbCollector_t *fbCollectorAllocTLS(
    fbListener_t    *listener,
    void            *ctx,
    int             fd,
    struct sockaddr *peer,
    size_t          peerlen,
    GError          **err)
{
    gboolean        ok = TRUE;
    fbCollector_t   *collector = NULL;
    fbConnSpec_t    *spec = fbListenerGetConnSpec(listener);
    
    /* Create a new collector */
    collector = g_slice_new0(fbCollector_t);
    
    /* Fill it in */
    collector->listener = listener;
    collector->ctx = ctx;
    collector->stream.fd = fd;
    collector->active = TRUE;
    collector->coread = fbCollectorReadTLS;
    collector->coclose = fbCollectorCloseTLS;
    if (peerlen) {
        memcpy(&(collector->peer.so), peer, 
               (peerlen > sizeof(collector->peer)) ? 
                    sizeof(collector->peer) : peerlen);
    }


    /* Do TLS accept atop opened socket */
    switch (spec->transport) {
    case FB_TLS_TCP:
        ok = fbCollectorOpenTLS(collector, err);
        break;
#if HAVE_OPENSSL_DTLS
    case FB_DTLS_UDP:
#if HAVE_OPENSSL_DTLS_SCTP
    case FB_DTLS_SCTP:        
#endif
        ok = fbCollectorOpenDTLS(collector, err);
    break;
#endif
    default:
        g_assert_not_reached();
    }

    /* Nuke collector on TLS setup error */
    if (!ok) {
        g_slice_free(fbCollector_t, collector);
        return NULL;
    }
        
    /* All done */
    return collector;
}

#endif /* HAVE_OPENSSL */

gboolean        fbCollectMessage(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    /* Ensure stream is open */
    if (!collector->active) return FALSE;

    /* Attempt to read message */
    if (collector->coread(collector, msgbase, msglen, err)) return TRUE;
    
    /* Read failure; signal error */
    return FALSE;
}

void          *fbCollectorGetContext(
    fbCollector_t   *collector)
{
    return collector->ctx;
}

int             fbCollectorGetFD(
    fbCollector_t   *collector)
{
    return collector->stream.fd;
}

void            fbCollectorClose(
    fbCollector_t   *collector)
{
    if (collector->active && collector->coclose) collector->coclose(collector);

    if (collector->listener) {
        fbListenerRemove(collector->listener, collector->stream.fd);
    }
}

void            fbCollectorFree(
    fbCollector_t   *collector)
{
    fbCollectorClose(collector);
    g_slice_free(fbCollector_t, collector);
}

